import 'dart:async';
import 'package:adeptannotations/adeptannotations.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:source_gen/source_gen.dart';

class CloneGenerator extends GeneratorForAnnotation<Cloneable> {
  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement = _getClassElement(element);
    String cloneFunction = _generateCloneFunction(classElement, annotation);
    return '''
    $cloneFunction
    ''';
  }

  final _cloneKeyAnnotationChecker = const TypeChecker.fromRuntime(CloneKey);
  final _cloneableKeyAnnotationChecker =
      const TypeChecker.fromRuntime(Cloneable);

  /// Check if the classElement has a default constructor like MyClass()
  _checkDefaultConstructor(ClassElement classElement) {
    bool hasDefaultConstructor = false;
    classElement.constructors?.forEach((constructor) {
      if (constructor?.isDefaultConstructor ?? false) {
        hasDefaultConstructor = true;
        return;
      }
    });
    if (!hasDefaultConstructor)
      throw InvalidGenerationSourceError(
          'To use @Cloneable make sure '
          'to have a default constructor in your class '
          '${classElement.displayName}.',
          todo: 'Add a datult constructor like MyClass() or '
              'MyClass({String param1, String param2}) to your '
              '${classElement.displayName} or remove the @Cloneable '
              'annotation from it.',
          element: classElement);
  }

  /// Get ClassElement from generic Element if possible, otherwise throws
  /// exception
  ClassElement _getClassElement(Element element) {
    if (element is! ClassElement)
      throw InvalidGenerationSourceError(
          'Generator cannot target '
          '${element.displayName}. @Cloneable can only be applied to a '
          'class.',
          todo: 'Check the type of ${element.displayName} or remove the '
              'annotation from it.',
          element: element);
    return element as ClassElement;
  }

  /// This function manages the logic of generating the Clone boilerplate code
  String _generateCloneFunction(
      ClassElement classElement, ConstantReader annotation) {
    _checkDefaultConstructor(classElement);
    String classType = classElement?.displayName;
    String functionName = "\$${classType}Clone";
    String currentDataParam = "currentData";
    String newDataParam = "newData";
    String mergeParam = "merge";
    String functionParams = "{$classType $currentDataParam, "
        "$classType $newDataParam, bool $mergeParam = false}";
    String cloneVar = "clone";

    StringBuffer cloneData = StringBuffer("final $cloneVar = $classType()");
    StringBuffer mergeData = StringBuffer("if($mergeParam ?? false){");
    Cloneable cloneable = _getCloneable(classElement) ?? Cloneable();
    classElement?.fields?.forEach((field) {
      if (field.isSynthetic && cloneable.ignoreSynthetic) return;
      if (field.isFinal || field.isConst || field.isStatic) return;
      if (field.isSynthetic && field.getter == null) return;
      CloneKey cloneKey = _getCloneKey(field);
      if (cloneKey?.ignore ?? false) return;
      String fieldName = field?.displayName;
      String cloneDataFieldValue = "$newDataParam?.$fieldName ?? "
          "${currentDataParam}.$fieldName";
      if (cloneKey?.forceNewData ?? false)
        cloneDataFieldValue = "$newDataParam?.$fieldName";
      cloneData.writeln("..${fieldName} = $cloneDataFieldValue");
      mergeData
          .writeln("${currentDataParam}.$fieldName = $cloneVar?.$fieldName;");
    });
    cloneData.writeln(";");
    mergeData.writeln("}");

    return '''
      /// This function creates a new instance of [$classType] by creating a 
      /// copy of [$currentDataParam] and replacing its fields with the data in 
      /// [$newDataParam].
      /// Notes:
      /// 1. The function will use the values from [$newDataParam], in case of 
      /// null then $currentDataParam values will be used. 
      /// 2. If [$mergeParam] is true then [$currentDataParam] will mirror the  
      /// new instance returned.
      /// 3. If @CloneKey(forceNewData: true) annotation is used in some field, 
      /// then [$newDataParam] value will be forced for the specified field 
      /// no matter it's nullability 
      $classType $functionName($functionParams){
        ${cloneData.toString()}
        ${mergeData.toString()}
        return $cloneVar;
      }
    ''';
  }

  /// Gets the Cloneable annotation for a given ClassElement
  Cloneable _getCloneable(ClassElement classElement) {
    DartObject cloneableAnnotation = _cloneableKeyAnnotationChecker
        .firstAnnotationOf(classElement, throwOnUnresolved: false);
    if (cloneableAnnotation != null)
      return Cloneable(
          ignoreSynthetic:
              cloneableAnnotation?.getField('ignoreSynthetic')?.toBoolValue());
    return null;
  }

  /// Gets the CloneKey annotation for a given FieldElement
  CloneKey _getCloneKey(FieldElement field) {
    Iterable<DartObject> annotations =
        _cloneKeyAnnotationChecker.annotationsOfExact(field);
    int annotationsCount = annotations?.length ?? 0;
    if (annotationsCount > 1)
      throw InvalidGenerationSourceError(
          'Repeated `@CloneKey` annotation on ${field?.displayName}.',
          todo: 'Check the `@CloneKey` annotations on field '
              '${field?.displayName}.',
          element: field);
    if (annotationsCount == 1) {
      DartObject annotation = annotations.first;
      return CloneKey(
        ignore: annotation.getField('ignore')?.toBoolValue(),
        forceNewData: annotation.getField('forceNewData')?.toBoolValue(),
      );
    }
    return null;
  }
}
